# -*- Mode: Perl -*-

###############################################################
#
#      File: Grammar/ContextFree/Generator.pm
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: GPL
#
# Description: simple perl OO Grammar
# 
###############################################################

package Grammar::ContextFree::Generator;
# System modules
use Carp;
use FileHandle;
# third party extensions
#use Set::Scalar;
# user extension modules
use Grammar::Tree;
#use Grammar::ContextFree::Rule::Simple;
#use Grammar::ListUtils;
#use Grammar::StringUtils;
@ISA       = qw();
@EXPORT    = qw();
@EXPORT_OK = qw();

###############################################################
# Package-global variables
###############################################################
$VERSION = 0.02;
sub VERSION {
  my ($self) = shift;
  return $self->{VERSION} ? $self->{VERSION} : $VERSION;
}

$Debug = 0;

###############################################################
# Constructor
#   : A grammar is a list-reference:
#     [ rules, startsym, epsilon, eoi ]
#   : rules is a hashref of the form
#     { lhs1 -> [ rhs1-1, ... , rhs1-J ],
#       ...
#       lhsN -> [ rhsM-1, ... , rhsN-K ] }
###############################################################
sub new {
  my $that = shift;
  my $self = {
	      rules=>{},             # rules
	      #Set::Scalar->new(),   # nonterminals
	      start=>'S',            # startsymbol
	      epsilon=>'',           # epsilon
              eoi=>"\$",             # eoi
	      maxexpansions=>undef, # max expansions (hack: undef=none)
	      expand_dummy=>['...'], # dummy expansion value
	     };
  my $class = ref($that) || $that;
  bless $self, $class;
  return $self->initialize(@_);
}

###############################################################
# Initialization
###############################################################
sub initialize {
  my ($self, %args) = @_;
  my ($key, $val, $code);
  while (($key, $val) = each(%args)) {
    if ($code = $self->can($key)) {
      &$code($val);
    }
    else {
      warn("Unknown key to ", ref($self), ": $key");
    }
  }
  return $self;
}

###############################################################
# Accessors / manipulators
#   rules       : hashref
#   startsymbol : scalar
#   epsilon     : scalar
#   eoi         : scalar
#   maxexpansions : scalar
###############################################################
sub rules {
  my $self = shift;
  # set value
  if (@_) { $self->{rules} = shift; } 
  # get value
  return $self->{rules};
}
sub startsymbol {
  my $self = shift;
  # set value
  if (@_) { $self->{start} = shift; } 
  # get value
  return $self->{start};
}
sub epsilon {
  my $self = shift;
  # set value
  if (@_) { $self->{epsilon} = shift; } 
  # get value
  return $self->{epsilon};
}
sub eoi {
  my $self = shift;
  # set value
  if (@_) { $self->{eoi} = shift; } 
  # get value
  return $self->{eoi};
}
sub maxexpansions {
  my $self = shift;
  # set value
  if (@_) { $self->{maxexpansions} = shift; } 
  # get value
  return $self->{maxexpansions};
}


###############################################################
# is_terminal($symbol)
# is_nonterminal($symbol)
# is_startsymbol($symbol)
# is_epsilon($symbol)
#   + test predicates
###############################################################
sub is_terminal {
  my ($self, $symbol) = @_;
  return (defined($symbol) && 
	  ! exists($self->{rules}->{$symbol}));
}
sub is_nonterminal {
  my ($self, $symbol) = @_;
  return (defined($symbol) && 
	  exists($self->{rules}->{$symbol}));
}
sub is_startsymbol {
  my ($self, $symbol) = @_;
  return (defined($symbol) && $symbol eq $self->{start});
}
sub is_epsilon {
  my ($self, $symbol) = @_;
  return (defined($symbol) && $symbol eq $self->{epsilon});
}

sub nonterminals { # returns a list of all nonterminals
  my $self = shift;
  return keys(%{$self->{rules}});
}

###############################################################
# clear
#   + removes all information stored in the grammar
###############################################################
sub clear {
  my $self = shift;
  %{$self->{rules}} = ();
}

###############################################################
# add_rule_from_string($string)
#   + adds the rule parsed from $string to the grammar,
#     where $string has the form:
#    "lhs -> rhs"
###############################################################
sub add_rule_from_string {
  my ($self, $rulestring) = @_;  
  if ($rulestring =~ /^\s*(\S+)\s*->\s*(.*\S)\s*$/) {
    # add information
    if ($self->is_nonterminal($1)) {
      # we already know about it -- just add an expansion
      push(@{$self->{rules}->{$1}}, $2);
    }
    else {
      # add nonterminal and make a new expansion-list
      $self->{rules}->{$1} = [ $2 ];
    }
  } 
  else { 
    chomp($rulestring);
    warn("Invalid rule specification string \"$rulestring\"");
  }
  return $self;
}

###############################################################
# read_rules_filehandle($FileHandle)
#   + reads in rulestrings from FileHandle, ignoring blank
#     lines and comment-lines (lines whose first non-space
#     character is "#")
###############################################################
sub read_rules_filehandle {
  my ($self, $fh) = @_;
  unless (defined($fh)) { 
    warn("bad filehandle for read"); 
    return undef; 
  }
  my $line;
  while (defined($fh) && ($line = <$fh>)) {
    next if ($line =~ /^\s*(?:\#.*)?$/); # ignore blank lines and comments
    $self->add_rule_from_string($line);
  }
  return 1;
}


###############################################################
# read_rules_file(@filenames)
#   + reads in rulestrings from files in @filenames, ignoring
#     lines and comment-lines (lines whose first non-space
#     character is "#")
###############################################################
sub read_rules_file {
  my $self = shift;
  my ($filename);
  while ($filename = shift) {
    my $fh = new FileHandle("<$filename");
    unless (defined($fh)) {
      warn("could not open \"$filename\" for reading: $!");
      return undef;
    }
    $self->read_rules_filehandle($fh);
    $fh->close;
  }
  return 1;
}

###############################################################
# merge($grammar)
#   + merges in the information from $grammar
###############################################################
sub merge {
  my ($self,$other) = @_;
  my ($lhs,$exps,$myexps);
  while (($lhs,$exps) = each(%{$other->rules()})) {
    $myexps = $self->{rules}->{$lhs};
    unless (defined($myexps)) {
      $self->{rules}->{$lhs} = [ @$exps ];
    } else {
      push(@$myexps, @$exps);
    }
  }
}


###############################################################
# compile([@filenames])
#    + extracts nonterminals & terminals from rules
#    + with @filenames, reads rules from @filenames first
#    + simple alias
###############################################################
*compile = \&read_rules_file;

###############################################################
# expansions($nonterminal)
# compute_expansions($nonterminal)
#    + returns a reference to a list of expansions
#      (themselves either scalars or listrefs) whose
#      lhs is eq to $nonterminal
###############################################################
sub expansions {
  my ($self,$nonterm) = @_;
  return $self->{rules}{$nonterm};
}

###############################################################
# expand_tree($tree)
#   + randomly expands all nonterminals in $tree
#   + returns the expanded $tree
#   + $tree is a Grammar::Tree object
###############################################################
sub expand_tree {
  my ($self, $t) = @_;
  unless (defined($t)) {
    $t = Grammar::Tree->new();
    $t->root_label($self->{start});
  }
  my ($sym, $exps, $exp, $h);
  my $expanded = 0;
  while ( !defined($h) || ($h = $t->next_topdown($h))) {
    $h = $t->root_node() if (!defined($h));
    my $sym = $t->label($h);
    if ($self->is_nonterminal($sym) && $t->is_leaf_node($h)) {
      ##-- unexpanded nonterminal

      ##-- throttle if we've already expanded too much...
      if (defined($self->{maxexpansions})
	  && $expanded >= $self->{maxexpansions}) {
	$t->add_children([@{$self->{expand_dummy}}], $h);
      }
      else {
	#-- do an expansion
	$expanded++;
	$exps = $self->expansions($sym);
	$exp = int(rand($#$exps + 1));

	#splice(@$l, $i, 1, split(/\s+/, $exps->[$exp]));
	$t->add_children([split(/\s+/, $exps->[$exp])], $h);

	if ($Debug) {
	  print STDERR
	    "[$nonterm -> $exps->[$exp]] => ", 
	      $t->canonical_string(), 
		"\n";
	}
      }
    }
    # must be a terminal -- keep looking
  }
  return $t;
}


###############################################################
# expand_list(\@listref)
#   + \@listref is a list of symbols, which is expanded
#     until it contains no more nonterminals
#   + returns a Grammar::Tree object
###############################################################
sub expand_list {
  my ($self, $l) = @_;
  my $tree = Grammar::Tree->new();
  if ($#$l > 0) {
    $tree->root_label('LIST');
    $tree->add_children($l);
  } else {
    $tree->root_label($l->[0]);
  }
  return $self->expand_tree($tree);
}

1;
