# -*- Mode: Perl -*-

###############################################################
#
#      File: Grammar::Tree.pm
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: GPL
#
# Description: build syntax trees for perl parsers
# 
###############################################################

package Grammar::Tree;
# System modules
use FileHandle;
use Carp;
require Exporter;
# 3rd party exstensions
#use Tree::MultiNode;
# user extension modules
#use StringUtils;
#use Gensym::String;
#use Populater::Child;
@ISA       = qw();
@EXPORT    = qw();
@EXPORT_OK = qw();

$NodeNum = 0;

###############################################################
# Package-global variables
###############################################################
our $VERSION = 0.01;
sub VERSION {
  my ($self) = shift;
  return $self->{VERSION} ? $self->{VERSION} : $VERSION;
}

our $Tk_SkipX = 15;
our $Tk_SkipY = 35;
our $Tk_LinePad = 2;
our $Tk_Font = 'helvetica -12 bold';
our $Tk_BoxActive = 1;

###############################################################
# Constructor
#   + structure: hash-ref
#     {
#      moms=>{$key=>$momkey,...},
#      #index => {$key=>$index,...},
#      dtrs =>{$key=>\@daughters,...},
#      #dindex => {$key=>$index},
#      labels=>{$key=>$label,...},
#      #current=>$node,
#      epsilon=>$epsilon,
#     }
###############################################################
sub new {
  my $that = shift;
  my %args = @_;
  return $that->copy(%args) if (ref($that));

  my ($label,$key) = @args{qw(label key)};
  $key = '_' if (!defined($key));

  my $self = {
	      key => $key,
	      labels => {$key=>$label}, # node labels
	      moms => {$key=>$mom},    # node moms (keys)
	      dtrs => {},   # daughters' keys
	      #index => {},  # node index in mom's 'dtrs' array
	      dindex => {}, # current daughter-indices
	      epsilon => '', # epsilon-marker (string)
	      (ref($that) ? (%$that) : qw()),
	      %args,
	     };
  return bless $self, ref($that) || $that;
}
sub copy {
  my $that = shift;
  my %args = @_;
  my $this = 
    {
     %$that,
     labels => { %{$that->{labels}}, },
     moms => { %{$that->{moms}}, },
     dtrs => { %{$that->{dtrs}}, },
     dindex => { %{$that->{dtrs}}, },
    };
  return bless $this, ref($that);
}

# ???
sub DESTROY {
  my $self = shift;
  #print STDERR __PACKAGE__, "::DESTROY() called.\n";
  $self->cleartree();
  %$self = ();
}

###############################################################
# Basic accessors
###############################################################
sub root_node {
  my $t = shift;
  return $t->{key};
}

sub root_label {
  my $t = shift;
  my $k = $t->root_node();
  $t->{labels}{$k} = shift(@_) if (@_);
  return $t->{labels}{$k};
}

sub label {
  my ($t,$n) = (shift,shift);
  $t->{labels}{$n} = shift(@_) if (@_);
  return $t->{labels}{$n};
}

sub children {
  my ($t,$n) = (shift,shift);
  if (defined($t->{dtrs}{$n})) { return @{$t->{dtrs}{$n}}; }
  return qw();
}

###############################################################
# clear
#   + clears the contents of the parse tree
###############################################################
*clear = \&cleartree;
sub cleartree {
  my $t = shift;
  $t->{key} = '_' if (!defined($t->{key}));
  %{$t->{labels}} = ($t->{key}=>'');
  %{$t->{moms}} = ($t->{key}=>undef);
  %{$t->{dtrs}} = ();
  %{$t->{dindex}} = ();
  return $t;
}


###############################################################
# $tree->get_key($node);
#   + gets tree-domain key
###############################################################
sub get_key {
  my ($tree,$node) = @_;
  return defined($node) ? $node : '';
}

###############################################################
# $tree->is_leaf_node($node)
#   + returns true if $node has no children
#   + allows non-method invocation (can call without $self
#     argument)
###############################################################
sub is_leaf_node {
  my ($tree,$node) = @_;
  $node = $tree->{key} if (!defined($node));
  return !defined($tree->{dtrs}{$node}) || $#{$tree->{dtrs}{$node}} < 0;
}

###############################################################
# $tree->is_root_node($node)
#   + returns true if $handle's depth is 0
#   + allows non-method invocation (can call without $self
#     argument)
###############################################################
sub is_root_node {
  my ($tree,$node) = @_;
  $node = $tree->{key} if (!$node);
  return !defined($tree->{moms}{$node});
}

###############################################################
# $tree->is_current_node($node)
#   + returns true if $handle's key is eq to the key of
#     $self->{handle}
#   + allows method-invocation only!
###############################################################
sub is_current_node {
  my ($t,$n) = @_;
  $n = $t->{key} if (!$n);
  return 1 unless (defined($n));
  return defined($t->{current}) && $t->{current} eq $h;
}

###############################################################
# $kidid = $tree->add_child_node($label,$mom);
#   + add a new child node under $mom
###############################################################
sub add_child_node {
  my ($t,$lab,$mom) = @_;
  $mom = $t->{key} if (!$mom);
  $t->{dtrs}{$mom} = [] if (!$t->{dtrs}{$mom});
  my $i = $#{$t->{dtrs}{$mom}}+1;
  my $k = $mom.'.'.$i;
  push(@{$t->{dtrs}{$mom}}, $k);
  $t->{labels}{$k} = $lab;
  $t->{moms}{$k} = $mom;
  return $k;
}

###############################################################
# $t = $t->add_children(\@labels,$mom)
###############################################################
sub add_children {
  my ($t,$labs,$mom) = @_;
  $mom = $t->{key} if (!$mom);
  foreach (@$labs) { $t->add_child_node($_,$mom); }
  return $t;
}


###############################################################
# $next = $t->next_topdown($node)
#   + returns the next (top-down, depth-first) node
#   + first checks node's next daughter position; if
#     that fails, there are no more child positions
#     available under $node.
#   + otherwise, checks mother instead -- if undef is
#     returned, $node is the root of the tree, and we
#     can return undef ourselves.
#   + otherwise, the last 2 steps are repeated until some
#     defined next-daughter is found
#     - An exception is the case where the next defined
#       child node's value is $self->{epsilon} -- this causes
#       resumption of the loop just as if no more children
#       were available.
#   + If some defined daughter is found, she is returned.
###############################################################
sub next_topdown {
  my ($t,$node) = @_;
  $node = $t->{key} if (!$node);
  my ($pos,$next);
 FINDKID: {
    do {
      # try getting next daughter
      $pos = $t->{dindex}{$node}++;
      if (defined($pos)
	  && defined($t->{dtrs}{$node})
	  && defined($next = $t->{dtrs}{$node}[$pos])) {
	# found an open child...
	$node = $next;
	if ($t->{labels}{$next} eq $t->{epsilon}) {
	  # ...bummer, it's epsilon.
	  ;
	} else {
	  # ...it's good
	  last FINDKID;
	}
      }
      # no more children: look at mother-node, if we have one
      return undef unless($node = $t->{moms}{$node});
    } while (1);
  }
  return $node;
}

###############################################################
# $next = $tree->expand(\@labels,$node)
#   + adds children in @$values and calls next_topdown($node)
###############################################################
sub expand {
  my ($t,$labels,$node) = @_;
  $node = $tree->{key} if (!$node);
  $t->add_children($labels,$node);
  return $t->next_topdown($node);
}

###############################################################
# \@leaves = $tree->leaf_list($node)
# \@leaves = $tree->leaf_list()
#   + returns a listref of leaves dominated by $node,
#     which defaults to $tree.
###############################################################
sub leaf_list {
  my ($t,$node) = @_;
  $node = $t->{key} if (!defined($node));
  my $ll = [];
  $t->traverse(node => $node,
	       sub => \&_leaf_list_node,
	       leaflist => $ll);
  return $ll;
}
sub _leaf_list_node {
  my (%args) = @_;
  if ($args{tree}->is_leaf_node($args{node})) {
    push(@{$args{leaflist}}, $args{tree}{labels}{$args{node}});
  }
}

###############################################################
# $tree->traverse(%args)
#    + abstract method for tree traversal, based on the example
#      code in the documentation of the Tree::MultiNode module.
#    + %args is a hash which may have the following keys:
#        tree    => $tree (caller)
#        sub     => \&sub,
#        after   => \&sub_after
#        node    => $node,
#        depth   => $depth,
#        first   => \&first,
#        last    => \&last,
#        butlast => \&butlast
#   + recursively descends the tree from $node (default is
#     $tree, calling &$sub(%args) for each
#     node.
#   + $sub_after(%args) is called after processing all children
#   + &first(%args) is called (if specified) for the first of 
#     a node's children, before calling traverse(%args) for that
#     child.
#   + &last(%args) is called (if specified) for the last of a 
#     node's children, after having called traverse(%args) for
#     that child
#   + &butlast(%args) is called (if specified) for all but the
#     last of a node's children, after having called
#     traverse(%args) for that child.
#   + The values returned by each call of any of the code
#     references \&sub, \&first, \&last, and \&butlast are.
#     gathered into a list which is returned by traverse
###############################################################
sub traverse {
  my ($tree,%args) = @_;
  # initialization
  $args{node} = $tree->{key} if (!defined($args{node}));
  $args{depth} = -1 if (!defined($args{depth}));
  $args{tree} = $tree if (!defined($args{tree}));
  ++$args{depth};
  my $node = $args{node};
  my (@list,$nkids,$i);

  # current node
  @list = &{$args{sub}}(%args) if (defined($args{sub}));
  $nkids = defined($tree->{dtrs}{$node}) ? $#{$tree->{dtrs}{$node}} : -1;
  for ($i = 0; $i <= $nkids; $i++) {
    # recursive transversal
    $args{node} = $tree->{dtrs}{$node}[$i];
    push(@list,
	 ($i == 0 && defined($args{first}) ? &{$args{first}}(%args) : qw()),
	 ($tree->traverse(%args)),
	 ($i != $nkids && defined($args{butlast}) ? &{$args{butlast}}(%args) : qw()),
	 ($i == $nkids && defined($args{last}) ? &{$args{last}}(%args) : qw()),
	);
  }
  --$args{depth};
  $args{node} = $node;
  # post-processing
  push(@list, &{$args{after}}(%args)) if (defined($args{after}));
  return @list;
}

###############################################################
# $tree->dump_tree();
# $tree->dump_tree($filehandle);
# $tree->dump_tree($filehandle,$node);
#   + top-level decomposed version of the example code in the
#     docs for Tree::MultiNode
#   + $node defaults to $tree
###############################################################
sub dump_tree { 
  my ($tree,$fh,$node) = @_;
  $fh = STDOUT if (!defined($fh));
  $node = $tree->{key} if (!$node);
  $fh->print($tree->dump_tree_list($node));
}

###############################################################
# $tree->dump_tree_list($node)
# $node->dump_tree_list()
#   + decomposed version of the example code in the
#     docs for Tree::MultiNode (depth-passing, undef checking)
#   + returns a list of strings which can be passed to "print"
#   + $node defaults to $tree
###############################################################
sub dump_tree_list {
  my ($tree,$node) = @_;
  $node = $tree->{key} if (!$node);
  return $tree->traverse(sub => \&dump_tree_node,
			 node => $node,
			 depth => -1);
}
sub dump_tree_node {
  my (%args) = @_;
  my $lead = ' ' x ($args{depth}*2);
  my $label = $args{tree}{labels}{$args{node}};
  #my ($key,$val) = $args{handle}->get_data();
  my @l = ($lead, "label:   ", defined($label) ? $label : "undef", "\n",
	   #$lead, "key:   ", defined($key) ? $key : "undef", "\n",
	   #$lead, "val:   ", defined($val) ? $val : "undef", "\n",
	   $lead, "depth: ", defined($depth) ? $args{depth} : "undef", "\n$lead--\n");
  return @l;
}

###############################################################
# $tree->string($string)
# $tree->string()
#   + sets tree structure to reflect that of $string, if
#     specified,
#   + otherwise, returns string rep of tree
###############################################################
sub string {
  my $tree = shift;
  return $tree->canonical_string() if (!@_);
  my $str = shift;
  $tree->clear;
  my $node = $tree->{key};
  while ($str) {
    if ($str =~ s/^([^<(,)>]+)//) { # symbol
      $tree->{labels}{$node} = $1;
    }
    #elsif ($str =~ s/^<([^<(,)>]+)>//) { # focussed node
    #  $node->{label} = $1;
    #  $self->{handle} = $handle->new($handle);  # hack!
    #}
    elsif ($str =~ s/^\(//) { # first daughter
      $node = $tree->add_child($node);
    }
    elsif ($str =~ s/^,//) {  # next daughter
      $node = $tree->{moms}{$node} if (defined($tree->{moms}{$node}));
      $node = $tree->add_child($node);
    }
    elsif ($str =~ s/^\)//) { # last daughter
      $node = $tree->{moms}{$node} if (defined($tree->{moms}{$node}));
    } else {
      warn ("cannot interpret tree-string from `$str'\n");
      last;
    }
  }
}

###############################################################
# $tree->canonical_string($node)
# $node->canonical_string()
#   + prints the list returned by tree_ascii_list as a string,
#     which constitutes an ASCII representation of the
#     (sub)tree rooted at $handle.
###############################################################
sub canonical_string { 
  my ($tree,$node) = @_;
  return join('',$tree->tree_ascii_list($node||$tree->{key}));
}

###############################################################
# draw_tree_tk(%args)
#   + %args keys:
#       canvas    : the Tk::Canvas object on which to draw the tree
#                   -> REQUIRED
#       handle    : the root of the (sub-)tree to draw
#                   -> defaults to a new handle into $tree at the root
#       xskip     : min horizontal space (pixels) between nodes
#                   -> defaults to $Grammar::Tree::Tk_SkipX
#       yskip     : vertical space (pixels) between nodes
#                   -> defaults to $Grammar::Tree::Tk_SkipY
#       linepad   : vertical space between node-center and edges
#                 -> defaults to $Grammar::Tree::Tk_LinePad
#       font      : text font
#                   -> defaults to $Grammar::Tree::Tk_Font
#       boxactive : boolean - whether to box the current handle
#                   -> $Grammar::Tree::Tk_BoxActive
#   + draws the the parse tree rooted at $handle 
#     on a Tk::Canvas object $canvas
###############################################################
sub draw_tree_tk {
  my ($tree, %args) = @_;
  unless (defined($args{canvas})) {
    warn("no canvas object to draw_tree_tk!\n");
    return undef;
  }
  eval "use Tk;";
  $args{xskip} = $Grammar::Tree::Tk_SkipX unless (defined($args{xskip}));
  $args{yskip} = $Grammar::Tree::Tk_SkipY unless (defined($args{yskip}));
  $args{linepad} = $Grammar::Tree::Tk_LinePad unless (defined($args{linepad}));
  $args{font} = $Grammar::Tree::Tk_Font unless (defined($args{font}));
  $args{boxactive} = $Grammar::Tree::Tk_BoxActive unless (defined($args{boxactive}));
  $args{node} = $tree->root_node() unless (defined($args{node}));

  my $pred = undef;
  $tree->traverse(%args,
		  sub => \&dtk_do_node,
		  after => \&dtk_on_up,
		  nodeinfo => {},
		  ancestors => [],
		  pred => \$pred,
		  #current => $self->{handle},
		  tags => [$tree],
		 );
  $tree->show_tree_tk(%args);
}

sub show_tree_tk {
  my ($tree, %args) = @_;
  # --- make sure the canvas is sane ---
  #$canvas->configure(-scrollregion => [$canvas->bbox('node', 'line')]);
  my @bbox = $args{canvas}->bbox("$tree");
  $args{canvas}->configure(-scrollregion => [
					     $bbox[0]-$args{xskip},
					     $bbox[1]-$args{yskip},
					     $bbox[2]+$args{xskip},
					     $bbox[3]+$args{yskip},
					    ]);
}

sub dtk_do_node {
  my (%args) = @_;
  $args{nodeinfo}->{$args{node}} =
    {
     #TreeNode => $args{handle}->{curr_node}, # undocumented Tree::MultiNode feature
     key => $args{node},
     label => $args{tree}->label($args{node}),
     depth => $args{depth},
     ancestors => [ @{$args{ancestors}} ],
     pred => defined($args{pred}) ? $$args{pred} : undef, 
    };
  push(@{$args{ancestors}}, "d".$args{node});
  # interior nodes are handled by dtk_on_up
  return qw() if (scalar($args{tree}->children($args{node})));
  dtk_draw_node(%args);
}
sub dtk_on_up { 
  my (%args) = @_;
  $$args{pred} = $args{node};
  pop(@{$args{ancestors}});
  return qw() if ($args{tree}->is_leaf_node($args{node}));
  dtk_draw_node(@_);
}

sub dtk_draw_node {
  my (%args) = @_;
  my $nodei = $args{nodeinfo}->{$args{node}};
  my $canvas = $args{canvas};
  my $id = $canvas->createText(0, 0, 
			       -tags => ['node',
					 $nodei->{key},
					 @{$nodei->{ancestors}},
					 @{$args{tags}},
					],
			       -text => $nodei->{label},
			       -font => $args{font} || '',
			      );
  $nodei->{id} = $id;

  # --- do the node placement ---
  # - preliminary node placement
  if (scalar($args{tree}->children($args{node}))) {
    # nonterminal node
    my @kbbox = $canvas->bbox("d" . $nodei->{key});
    $canvas->coords($id, ($kbbox[2] + $kbbox[0]) / 2, 0);
  }

  # - make sure it was ok -- catch terminals, too
  my @bbox = $canvas->bbox($nodei->{key}, "d".$nodei->{key});

  my ($xborder);
  my @pbbox = (defined($nodei->{pred})
	       ? $canvas->bbox($nodei->{pred}, "d".$nodei->{pred})
	       : qw());
  if (@pbbox) {
    $xborder = $pbbox[2] + $args{xskip};
    #print STDERR 
    #  "Huh?? node->value: $node->{value}, pred: $node->{pred}, xskip: $args{xskip}, pbbox[2] = $pbbox[2]\n";
  } else {
    $xborder = $args{xskip};
  }

  # x placement/adjustment
  if ($bbox[0] < $xborder) {
    # houston, we have a problem
    $canvas->move($nodei->{key}, $xborder - $bbox[0], 0);
    $canvas->move("d" . $nodei->{key}, $xborder - $bbox[0], 0);
  }

  # y placement/adjustment
  $canvas->move($nodei->{id}, 0, $args{yskip} + ($nodei->{depth} * $args{yskip}));

  # --- final height & width calculation ---
  @bbox = $canvas->bbox($nodei->{key});
  $nodei->{height} = $bbox[3] - $bbox[1];
  @bbox = $canvas->bbox($nodei->{key}, "d".$nodei->{key});
  $nodei->{width} = $bbox[2] - $bbox[0];

  # --- do box ? ---
  # include box ?
  if (defined($args{boxactive}) && defined($args{current}) && 
      $args{boxactive} && $args{node}->get_key eq $args{current}->get_key) {
    $canvas->createRectangle($canvas->bbox($id),
			     -tags => [$nodei->{key},
				       @{$nodei->{ancestors}},
				       @{$args{tags}}]);
  }

  # --- connect the daughters ---
  #my @dtrs = $node->{TreeNode}->child_keys();
  my @dtrs = $args{tree}->children($args{node});
  my ($dtrkey,$dtri,$dx,$dy,$mx,$my);
  foreach $dtrkey (@dtrs) {
    $dtri = $args{nodeinfo}->{$dtrkey};
    ($dx,$dy) = $canvas->coords($dtri->{id});
    ($mx,$my) = $canvas->coords($nodei->{id});
    $canvas->createLine($mx,
			$my + $args{linepad} + ($nodei->{height} / 2),
			$dx,
			$dy - ($args{linepad} + ($dtri->{height} / 2)),
			-tags => ["d".$nodei->{key}, @{$nodei->{ancestors}}, @{$args{tags}}]);
  }
}


###############################################################
# $node->print_tree_ascii()
# $node->print_tree_ascii($filehandle)
# $tree->print_tree_ascii($filehandle, $node])
#   + prints the list returned by tree_ascii_list, which
#     constitute an ASCII representation of the (sub)tree
#     rooted at $node
###############################################################
sub print_tree_ascii { 
  my ($tree,$fh,$node) = @_;
  $node = $tree->{key} if (!defined($node));
  $fh = STDOUT unless (defined($fh));
  $fh->print($tree->tree_ascii_list($node));
}

###############################################################
# $tree->tree_ascii_list($node)
# $node->tree_ascii_list()
#   + returns a list of strings which can be passed to "print"
#     and which constitute an ASCII representation of $handle,
#     and which bear a remarkable resemblance to a PROLOG term.
###############################################################
sub tree_ascii_list {
  my ($tree,$node) = @_;
  $node = $tree if (!defined($node));
  return $tree->traverse(sub => \&_tree_ascii_list_node,
			 first => \&_tree_ascii_list_first,
			 last => \&_tree_ascii_list_last,
			 butlast => \&_tree_ascii_list_butlast,
			 node => $node,
			 #current => $self->{current},
			);
}
sub _tree_ascii_list_node {
  my %args = @_;
  my $lab = $args{tree}{labels}{$args{node}};
  $lab = 'undef' unless (defined($lab));
  if (defined($args{current}) && $args{node} eq $args{current}) {
    return "<$lab>";
  }
  return $lab;
}
sub _tree_ascii_list_first { return '('; }
sub _tree_ascii_list_last { return ')'; }
sub _tree_ascii_list_butlast { return ','; }



###############################################################
# Autoload function allows
#  "$mytree->key" to get an arbitrary key value, and 
#  "$mytree->key(val)" to set the value of 'key' to 'val'.
###############################################################
sub AUTOLOAD {
  my $self = shift;
  my $type = ref($self) || croak "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*:://; # strip package-part of name
  if (exists($self->{$name})) {
    if (@_) { return $self->{$name} = shift; } # set value
    else { return $self->{$name}; }            # get value
  }
  carp(__PACKAGE__ , "::$name() method not found in AUTOLOAD");
  return undef;
}

1;
