###############################################################
#
#      File: SayWhat::Base
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: Perl Artistic License
#
# Description: Base functions for SayWhat generator
# 
###############################################################

package SayWhat::Base;

# system modules
use Carp;
use Exporter;
use POSIX qw(&isatty);

# user modules
use Grammar::ContextFree::Generator;
use SayWhat::Synthesizer qw(:gensym :rc);
#use SayWhat::Synthesizer::Rsynth;
use SayWhat::Synthesizer::Festival;

use Pd::Socket;

# 3rd-party modules
#use Speech::Festival;
#use Storable;

@ISA = qw(Exporter);

###############################################################
# Globals
###############################################################
our $VERSION = "0.18";
our $Prompt = 'SayWhat> ';

our $DEFAULT_SYNTH_TYPE = 'SayWhat::Synthesizer::Dummy';
our %DEFAULT_SYNTH_ARGS = (
			   SayWhat::Synthesizer => {},
			   SayWhat::Synthesizer::Rsynth => {},
			   SayWhat::Synthesizer::Festival => {},
			   SayWhat::Synthesizer::Dummy => {},
			  );
sub DEFAULT_SYNTH_TYPE { return $DEFAULT_SYNTH_TYPE; }
sub DEFAULT_SYNTH_ARGS { return \%DEFAULT_SYNTH_ARGS; }

our ($RC_FAILED,$RC_OK) = (0,1);

## %STATUS : default status messages
##   $op => [ [$VLEVEL_FAILED, $MSG_FAILED], [$VLEVEL_OK,$MSG_OK] ]
our %STATUS =
  (
   retrieve_grammar=>[[0,'Retrieve failed for stored grammar from `$_[0]\': $!'],
		      [1,'Retrieved stored grammar from `$_[0]\''],
		     ],
   load_grammar=>[[0,'Load failed for grammar from `$_[0]\': $!'],
		  [1,'Loaded grammar rules from `$_[0]\''],
		 ],
   store_grammar=>[[0,'Error storing grammar to `$_[0]\': $!'],
		   [1,'Stored grammar in `$_[0]\''],
		  ],
   add=>[[0,'add_utterance() failed'],
	 [1,'Added utterance \`$_[0]\''],
	],
   del=>[[0,'del_utterance() failed'],
	 [1,'Removed utterance \`$_[0]\''],
	],
   rename=>[[0,'rename_utterance() failed'],
	    [1,'Renamed utterance \`$_[0]\' -> \`$_[1]\''],
	   ],
   generate=>sub{
     my ($rv,$sw,$id,$str,$tree) = @_;
     $sw->vout(5, "-->generate status sub(", join(',', @_), ")\n");
     if ($rv) {
       $sw->vout(1,"$id: ");
       $sw->vout(0,"$str\n\n");
       return $id;
     } else {
       $sw->verr(0,"Error: generation failed!\n");
       return 0;
     }
   },
   eval_perl=>sub{
     print join(' ', 'eval_perl:', @_);
     return @_;
   }
  );

sub STATUS { return \%STATUS; }

###############################################################
# Constructor: new(%args)
#  + %args keys
#     wbound=>$str   : word-boundary string
#     grammar=>$g    : Grammar::ContextFree::Generator object
#     verbose=>$lvl  : verbosity level:
#                      -1 - silent
#                       0 - print only generated text
#                       1 - print status messages
#     synth=>$syn     : a SayWhat::Synthesizer object
#     synthtype=>$typ : type for synth
#     synthargs=>\%h  : args for new synth object (HASH-ref)
#     pd=>$bool       : whether to try pdsend() on generate() # OBSOLETE
#     pdsendfh=>$fh   : filehandle to pdsend()                # OBSOLETE
#     rc=>$bool       : last op status
#     op=>$op,        : last op name
#     postproc=>\&sub : string-postprocessing sub
#  + internal structure:
#     utts     : information on generated utterances
#                { alias => { text=>$string, tree=>$tree } }
#     _uttidx  : internal utterance-counting index
###############################################################
sub new {
  my ($that,%args) = @_;
  my $class = ref($that) || $that;
  my $self = bless {
		    wbound => ' ',
		    grammar => undef,
		    utts => {},
		    _uttidx => 0,
		    verbose => 1,
		    rc => 1,
		    op => 'create',
		    postproc=>\&default_postproc,
		    %args
		   }, $class;

  ##-- initialize grammar
  if (!defined($self->{grammar})) {
    $self->{grammar} = Grammar::ContextFree::Generator->new();
  }
  $self->{grammar}->maxexpansions($self->{maxexpansions})
    if (defined($self->{maxexpansions}));
  delete($self->{maxexpansions});

  ## initialize synth
  unless (UNIVERSAL::isa($self->{synth},'SayWhat::Synthesizer'))
    {
      $self->{synthtype} = $self->DEFAULT_SYNTH_TYPE() unless (defined($self->{synthtype}));
      unless (UNIVERSAL::isa($self->{synthtype},'SayWhat::Synthesizer')) {
	$self->{synthtype} = "SayWhat::Synthesizer::$self->{synthtype}";
      }

      my %synthargs =
	(
	 (defined($class->DEFAULT_SYNTH_ARGS->{$self->{synthtype}})
	  ? (%{$class->DEFAULT_SYNTH_ARGS->{$self->{synthtype}}})
	  : qw()),
	 (exists($self->{synthargs})
	  ? %{$self->{synthargs}}
	  : qw())
	);
      $self->{synth} = SayWhat::Synthesizer->new(type=>$self->{synthtype},
						 verbose=>$self->{verbose},
						 %synthargs);
    }
  delete($self->{synthargs});

  return $self;
}

sub DESTROY {
  %{$_[0]} = ();
}

###############################################################
# Verbosity-checking
###############################################################

# $sw->vout($level,@message)
#  + prints message @message to stdout if verbosity >= $level
sub vout {
  if ($_[0]->{verbose} >= $_[1]) {
    print STDOUT @_[2..$#_];
  }
}

# $sw->verr($level,@message);
#  + prints message @message to stderr if verbosity >= $level
sub verr {
  if ($_[0]->{verbose} >= $_[1]) {
    print STDOUT @_[2..$#_];
  }
}

# $sw->vdo($level,$CODE);
#  + evals $CODE if verbosity >= $level
sub vdo {
  if ($_[0]->{verbose} >= $_[1]) {
    eval $_[2];
  }
}

###############################################################
# Prompting
###############################################################

# $class->doprompt($lexwrapper);
#   + prints a prompt if $lexwrapper->file() returns a tty
sub doprompt {
  my ($that, $lexwrapper) = @_;
  my $class = ref($that)||$that;
  print ${"${class}::Prompt"} if (isatty($lexwrapper->file)
				  #|| $lexwrapper->file() eq 'STDIN'
				 );
}

###############################################################
# status
###############################################################

# undef = status($op,$rv,@args);
#  + show status for $op,
sub status {
  my ($sw,$op,$rv) = (shift,shift,shift);
  $sw->vout(4,
	    "status(): op=$op, rv=", (defined($rv) ? $rv : "undef"),
	    "args=",join(',',@_), "\n");
  my $rci = $rv ? 1 : 0;
  @$sw{qw(op rc)} = ($op,$rv);
  my ($status,$msg);
  my $STATUS = $sw->STATUS();
  if (defined($STATUS->{$op})) {
    if (ref($STATUS->{$op})) {
      if (ref($STATUS->{$op}) eq 'ARRAY') {
	if (defined($STATUS->{$op}[$rci])) {
	  $status = $STATUS->{$op}[$rci];
	  $msg = eval qq(qq($status->[1]));
	  $sw->vout($status->[0], $msg, "\n");
	  return $rv;
	} #/defined($STATUS->{$op}[$rci])
	##-- ignore undef rv-elements in extant message arrays
      }
      #/ref($STATUS->{$op}) eq 'ARRAY'
      elsif (ref($STATUS->{$op}) eq 'CODE') {
	return &{$STATUS->{$op}}($rv,@_);
      }
      #/ref($STATUS->{$op}) eq 'CODE'
    }
    #/ref($STATUS->{$op})
  }
  #/defined($STATUS->{$op})

  ##-- default
  $sw->vout(1, join(' ', $op.':', $rv, @_), "\n");
  return $rv;
}

###############################################################
# load_grammar($filename, [$mode])
#   + load rules-grammar
#   + $mode is either 'sto', 'rule', or 'auto'
###############################################################
sub load_grammar {
  my ($self,$filename,$mode) = @_;
  my $grammar = $self->{grammar};
  $mode = 'auto' unless defined($mode);
  my ($op,$rc) = qw(load_grammar 1);
  if ($mode eq 'sto' || ($mode ne 'rule' && $filename =~ /\.sto$/)) {
    eval "use Storable;";
    my $g = Storable::retrieve($filename);
    if ($g) {
      $grammar->merge($g);
    } else {
      $rc = 0;
    }
  }
  elsif (!$grammar->read_rules_file($filename)) {
    $rc = 0;
  }
  return $self->status($op,$rc,$filename);
}

sub store_grammar {
  my ($self,$filename) = @_;
  my $grammar = $self->{grammar};
  my ($op,$rv) = qw(store_grammar 1);
  eval "use Storable;";
  $rc = 0 if (!Storable::store($grammar,$filename));
  return $self->status($op,$rv,$filename);
}


###############################################################
# forget_grammar
#   + clear grammar
###############################################################
sub forget_grammar {
  $_[0]->{grammar}->clear;
  $_[0]->status('forget_grammar', 1);
}


###############################################################
# do_generate($nonterm) => [ $string, $tree ]
# do_generate($nonterm) => [ $string, $tree ]
#   + generates a phrase with root $nonterm
#   + the string generated is returned as $string
#   + the derivation-tree is returned as $tree
#   + $nonterm defaults to $self->{grammar}->startsymbol
###############################################################
sub do_generate {
  my ($self,$nonterm) = @_;
  my $grammar = $self->{grammar};
  $nonterm = $grammar->startsymbol unless (defined($nonterm));
  my @nonterms = split(/$self->{wbound}+/, $nonterm);
  my $tree = $grammar->expand_list(\@nonterms);
  my ($str);
  if (defined($self->{wbound}) && $self->{wbound} !~ /^\s+/) {
    $str = join('', @{$tree->leaf_list()});
    $str =~ s/$self->{wbound}/ /g;
  } else {
    $str = join(' ', @{$tree->leaf_list()});
  }
  if (defined($self->{postproc})) {
    $str = &{$self->{postproc}}($self,$str);
  }
  return [$str, $tree];
}

###############################################################
# undef = generate_post($str,$tree,$id)
#  + do random printing stuff
###############################################################
#sub generate_post {
#  my ($self,$str,$tree,$id) = @_;
#  $self->vout(1,"$id: ");
#  $self->vout(0,"$str\n\n");
#}

###############################################################
# $id = generate([$nonterm])
# $id = generate_say([$nonterm])
#   + calls generate($nonterm), and adds utterance-info to list
#   + prints string to stdout if verbose>=1
#   + second form pronounces string, as well
###############################################################
sub generate {
  my $self = shift;
  return $self->generate_to(undef,@_);
}
sub generate_say {
  my $self = shift;
  my $id = $self->generate(@_);
  $self->say_utterance($id);
  return $id;
}

###############################################################
# $id = generate_to($id,[$nonterm])
#   + calls generate($nonterm), and adds utterance-info for $id to list
#   + prints string to stdout if verbose>=1
###############################################################
sub generate_to {
  my $self = shift;
  my $id = shift || $self->gensym_utt();
  my $info = $self->do_generate(@_);
  $id = $self->add_utterance(@$info, $id);
  my $rv = defined($id) ? 1 : 0;
  #$self->generate_post(@$info,$id);
  return $self->status('generate',$rv,$self,$id,@$info) ? $id : undef;
}


###############################################################
# eval_perl($perlblock)
# eval_string($string)
#   + does what it says
###############################################################
sub eval_perl {
  my $sw = shift;
  my @rc = qw();
  if ($sw->{safe}) {
    $@ = __PACKAGE__ . "::eval_perl() not available in 'safe' mode!";
    @rc = qw('unavailable');
  } else {
    @rc = eval "package main; $_[0]";
  }
  print STDERR "$@\n" if ($@);
  return $sw->status('eval_perl',@rc);
}
sub eval_string {
  my ($sw,$q,$s) = @_;
  #print "eval_string(", join(',',@_), "\n";
  if ($sw->{safe}) {
    $s = '' if (!defined($s));
    $s =~ s/^$q//;
    $s =~ s/$q$//;
    $s =~ s/\\$q/$q/g;
    return $s;
  }
  my $val = eval "package main; $s";
  print STDERR "$@\n" if ($@);
  return $val;
}

###############################################################
# ll_utterances
#   + lists all known utterances to stdout (long format)x
###############################################################
sub ll_utterances {
  my $self = shift;
  my $regexp = shift;
  $regexp = '.' unless (defined($regexp));
  my $uttinfo = $self->{utts};
  local ($id,$str,$tree);
  my ($oldtop,$oldfmt,$oldformfeed) = ($^,$~,$^L);
  ($^,$~,$^L) = ('UTTLIST_TOP', 'UTTLIST', '');
  my @ids = sort(keys(%$uttinfo));
  print "Utterance-ID   Text                       Tree\n";
  print "(none)\n" unless (@ids);
  foreach $id (@ids) {
    #($str,$tree) = @{$uttinfo->{$id}}[1..2];
    ($str,$tree) = @{$uttinfo->{$id}}{qw(text tree)};
    $str = '' unless defined($str);
    unless (defined($tree) && ref($tree)) {
      $tree = '';
    } else {
      $tree = $tree->canonical_string;
    }
    next unless ($id =~ $regexp || $str =~ $regexp || $tree =~ $regexp);
    write;
  }
  ($^,$~,$^L) = ($oldtop,$oldfmt,$oldformfeed);
}
format UTTLIST_TOP =
.
format UTTLIST =
@<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$id,           $str,                      $tree
~~             ^<<<<<<<<<<<<<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
               $str,                      $tree
.

###############################################################
# ls_utterances
#   + lists all known utterances to stdout (short format)
###############################################################
*list_utterances = \&ls_utterances;
sub ls_utterances {
  my $self = shift;
  my $regexp = shift;
  $regexp = '.' unless (defined($regexp));
  my $uttinfo = $self->{utts};
  local ($id,$text);
  my ($oldtop,$oldfmt,$oldformfeed) = ($^,$~,$^L);
  ($^,$~,$^L) = ('UTTLS_TOP', 'UTTLS', '');
  my @ids = sort(keys(%$uttinfo));
  print "Utterance-ID   Text\n";
  print "(none)\n" unless (@ids);
  foreach $id (@ids) {
    $text = $uttinfo->{$id}{text};
    $text = '' unless defined($text);
    next unless ($id =~ $regexp || $text =~ $regexp);
    write;
  }
  ($^,$~,$^L) = ($oldtop,$oldfmt,$oldformfeed);
}
format UTTLS_TOP =
.
format UTTLS =
@<<<<<<<<<<<<< ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$id,           $text
~~             ^<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
               $text
.


###############################################################
# Synthesizer Type & Configuration
###############################################################

# $type = $sw->synth_type();
# $type = $sw->synth_type($type);
#  + get/set synth type
sub synth_type {
  my $sw = shift;

  ## get type
  return ref($sw->{synth}) unless (@_);

  ## set type: canonicalize
  my $type = shift;
  eval "require $type";
  eval "require SayWhat::Synthesizer::$type";
  unless (UNIVERSAL::isa($type,'SayWhat::Synthesizer')
	  ||
	  (($type = "SayWhat::Synthesizer::$type")
	   &&
	   UNIVERSAL::isa($type,'SayWhat::Synthesizer')))
    {
      carp("Error: '$type' is not a SayWhat::Synthesizer subtype!");
      return $sw->{synthtype};
    }

  ## don't re-create if redundant
  unless (ref($sw->{synth}) && ref($sw->{synth}) eq $type) {
    my $class = ref($sw) || $sw;
    $sw->{synth} =
      SayWhat::Synthesizer->new(type=>$type,
				verbose=>$sw->{verbose},
				(%{$class->DEFAULT_SYNTH_ARGS->{$sw->{synthtype}}}),
			       );
    ## perpetuate utterances
    my ($id,$utt);
    while (($id,$utt) = each(%{$sw->{utts}})) {
      $sw->{synth}->add_utterance($id,$utt->{text});
    }
  }

  return $sw->status('synthtype',$sw->{synthtype} = $type);
}

# %config = $sw->synth_config();
# $val = $sw->synth_config($key);
# %config = $sw->synth_config(%config);
sub synth_config {
  return shift->synth_wrapper('configure',@_);
}

# $sw->show_synth_config
#   + prints synth-configuration to stdout
sub show_synth_config {
  my $sw = shift;
  my %config = $sw->{synth}->configure();
  my ($oldtop,$oldfmt,$oldformfeed) = ($^,$~,$^L);
  ($^,$~,$^L) = ('CONFIG_TOP', 'CONFIG', '');
  local ($key,$val);
  print "                         OPTION = VALUE\n";
  if (@_) {
    $key = shift;
    $val = $config{$key};
    write;
  } else {
    foreach $key (sort(keys(%config))) {
      $val = $config{$key};
      write;
    }
  }
  ($^,$~,$^L) = ($oldtop,$oldfmt,$oldformfeed);
}
format CONFIG_TOP =
.
format CONFIG =
@>>>>>>>>>>>>>>>>>>>>>>>>>>>>>> = @<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
$key,                             $val
.


###############################################################
# $val = $sw->synth_wrapper($methodname,@args)
#  + wraps synth-method $methodname
###############################################################
sub synth_wrapper {
  my ($sw,$method,@args) = @_;
  if (!defined($sw->{synth})) {
    carp(ref($sw).": no synthesizer defined for synth_wrapper($method,...)\n");
    return undef;
  }
  return $sw->{synth}->$method(@args);
}


###############################################################
# subs: utterance maintainence
#   + $id = gensym_utt()
#   + $id = add_utterance($text[,$tree[,$id]]);  # CHANGED RETURN CONVENTION!
#   + $id = rename_utterance($oldid,$newid);     # CHANGED RETURN CONVENTION!
#   + $rc = del_utterance($id);
#   + $rc = synth_utterance($id);
#   + $rc = say_utterance($id);
#   + $rc = save_utterance($id,[$file,[$mode]]);
###############################################################

###############################################################
# add/del/rename
sub add_utterance {
  my ($self,$text,$tree,$id) = @_;

  ## setup tree
  $tree = '' unless (defined($tree));
  if ($tree && ! ref($tree)) {
    my $treestr = $tree;
    $tree = Grammar::Tree->new();
    $tree->string($treestr);
  }
  ## setup id
  $id = $self->gensym_utt() unless (defined($id));

  ## delete it if it already exists
  $self->del_utterance($id) if (exists($self->{utts}{$id}));

  ## send it to the synth
  return undef unless ($self->{synth}->add_utterance($id,$text) eq $RC_OK);

  ## add it ourselves
  $self->{utts}{$id} = {
			name=>$id,
			text=>$text,
			tree=>$tree,
		       };
  return $self->status('add', $id||0, $id);
}
sub del_utterance {
  my ($self,$id) = @_;
  return $RC_FAILED
    unless ($self->{synth}->del_utterance($id) eq $RC_OK);
  delete($self->{utts}{$id});
  return $self->status('del', 1, $id);
}

sub rename_utterance {
  my ($self,$oldid,$newid) = @_;

  return $RC_FAILED
    unless ($self->{synth}->rename_utterance($oldid,$newid) eq $RC_OK);

  my $utt = $self->{utts}{$oldid};
  if (!defined($utt)) {
    carp(ref($self)."->rename_utterance(): warning: utterance-id '$oldid' not defined!\n");
    return $RC_FAILED;
  }
  $self->{utts}{$newid} = $self->{utts}{$oldid};
  delete $self->{utts}{$oldid};
  return $self->status('rename',$newid,$oldid,$newid);
}

###############################################################
# synth/say/save
sub synth_utterance {
  my $sw = shift;
  return $sw->status('synth',$sw->synth_wrapper('synth_utterance',@_));
}
sub say_utterance {
  my $sw = shift;
  $sw->synth_wrapper('synth_utterance',@_);
  return $sw->status('say',$sw->synth_wrapper('say_utterance',@_));
}
sub save_utterance {
  my $sw = shift;
  return $sw->status('save',$sw->synth_wrapper('save_utterance',@_));
}

###############################################################
# save_script($filename)
#   + saves current state ($self->{utts}) as a script
#     to $filename
###############################################################
sub save_script {
  my ($self,$filename) = @_;
  my $outfh = FileHandle->new(">$filename");
  if (!defined($outfh)) {
    return $sw->status('save_script',0,$filename,$!);
  }
  $outfh->print("# SayWhat version $VERSION script file auto-generated by $0 ($$)\n\n",
		"# -- Globals\n",
		"set-wbound '$self->{wbound}'\n",
		"\n",
		"# -- Synthesizer\n",
		# add synth-type and configuration
		"synth-type ", ref($self->{synth}), "\n",
		"synth-config\\\n\t",
	       );
  my %config = $self->{synth}->configure();
  $outfh->print(join(",\\\n\t",
		     map {
		       "'$_' = '$config{$_}'"
		     } sort(keys(%config))),
		"\n\n",
		# add utterances
		"# -- Utterances\n",
	       );
  my $utts = $self->{utts};
  my ($id,$text,$tree);
  foreach $id (sort(keys(%$utts))) {
    ($text,$tree) = @{$utts->{$id}}{qw(text tree)};
    $tree = $tree->string if (ref($tree));
    $text = '' unless defined($text);
    $tree = '' unless defined($tree);

    $text =~ s/([\'])/\\$1/g;
    $tree =~ s/([\'])/\\$1/g;

    $outfh->print("add\\\n\t'$text'\\\n\t'$tree'\\\n\t'$id'\n\n");
  }
  $outfh->close;
  return $sw->status('save_script',1,$filename);
}

###############################################################
# Default postprocesing sub
###############################################################
# $str = postproc($sw,$str);
sub default_postproc {
  my ($sw,$s) = @_;
  $s =~ s/\b(a)\s+([aeiouAEIOU])/an $2/g;
  $s =~ s/\b(an)\s+([^aeiouAEIOU])/a $2/g;
  $s =~ s/^(.)/\u$1/;
  return $s;
}


###############################################################
# Autoload function allows
#  "$saywhat->key" to get an arbitrary key value, and 
#  "$saywhat->key(val)" to set the value of 'key' to 'val'.
#
# Passes unknown methods through to {synth} or {grammar}
###############################################################
sub AUTOLOAD {
  my $self = shift;
  my $type = ref($self) || croak "$self is not an object";
  my $name = $AUTOLOAD;
  $name =~ s/.*:://; # strip package-part of name
  unless (exists $self->{$name}) {
    if (defined($self->{synth}) && $self->{synth}->can($name)) {
      # pass unknown methods first to {synth}
      return $self->{synth}->$name(@_);
    } elsif (defined($self->{grammar}) && $self->{grammar}->can($name)) {
      # ... and try {grammar}, if that doesn't work
      return $self->{grammar}->$name(@_);
    }
    else {
      carp("Unknown method '$name' for object $self");
    }
  }
  if (@_) { return $self->{$name} = shift; } # set value
  else { return $self->{$name}; }            # get value
}



1;
