###############################################################
#
#      File: SayWhat::PD
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: Perl Artistic License
#
# Description: PD net interface for SayWhat generator
#
###############################################################

package SayWhat::PD;

# system modules
use POSIX qw(mkfifo);
use IO::File;

# user modules
use SayWhat::Base;


@ISA = qw(SayWhat::Base);

our $VERSION = 0.02;

our $Prompt = 'PDSayWhat> ';

#our $DEFAULT_FIFO = '/tmp/pd-saywhat.fifo';
our $DEFAULT_FIFO = undef;

our $DEFAULT_SYNTH_TYPE = __PACKAGE__->SUPER::DEFAULT_SYNTH_TYPE();
our %DEFAULT_SYNTH_ARGS = (
			   SayWhat::Synthesizer::Festival => {
							      mode => 'raw',
							      samp_rate => 44100,
							     },
			   SayWhat::Synthesizer::Rsynth => {
							    mode => 'raw',
							    samp_rate => 44100,
							   },
			  );

sub DEFAULT_SYNTH_TYPE { return $DEFAULT_SYNTH_TYPE; }
sub DEFAULT_SYNTH_ARGS { return \%DEFAULT_SYNTH_ARGS; }


###############################################################
# Constructor: new(%args)
#  + %args keys
#    NEW
#     pdfifo=>$path  : pathname to fifo (default: (none))
#    INHERITED
#     wbound=>$str   : word-boundary string
#     grammar=>$g    : Grammar::ContextFree::Generator object
#     verbose=>$lvl  : verbosity level:
#                      -1 - silent
#                       0 - print only generated text
#                       1 - print status messages
#     synth=>$syn     : a SayWhat::Synthesizer object
#     synthtype=>$typ : type for synth
#     synthargs=>\%h  : args for new synth object (HASH-ref)
#  + internal structure:
#     utts     : information on generated utterances
#                { alias => { text=>$string, tree=>$tree } }
#     _uttidx  : internal utterance-counting index
###############################################################
sub new {
  my ($that,%args) = @_;

  my $fifo = $args{pdfifo};
  delete($args{pdfifo});

  my $self = $that->SUPER::new(%args);
  bless $self, ref($that) || $that;

  ## setup fifo file
  $fifo = $DEFAULT_FIFO unless (defined($fifo));
  if (defined($fifo)) {
    unless (-p "$fifo") {
      mkfifo($fifo,0600) or die("$0: mkfifo failed for '$fifo': $!");
    }
  } else {
    $self->{fifo} = undef;
  }
    $self->{fifo} = $fifo;

  return $self;
}

sub DESTROY {
  # unlink fifo
  unlink($_[0]->{fifo}) if (defined($_[0]->{fifo}));
  # remove refs
  %{$_[0]} = ();
}


###############################################################
# Prompting & Messages
###############################################################

# $class->doprompt($lexwrapper);
#   + prints a prompt if $lexwrapper->file() returns a tty
#sub doprompt { ; }


##---------------------------------------------------------------
## @tokens = $sw->pd_sanitize_tokens(@untokenized_strings)
##  + tokenize @untokenized_strings
sub pd_sanitize_tokens2 {
  my $sw = shift;
  return
    (map {
      s/[\;\,]/ /g;
      (map { "t $_" } split(/\s+/, $_))
    } @_);
}
sub pd_sanitize_tokens {
  my $sw = shift;
  return
    (map {
      s/[\;\,]/ /g;
      split(/\s+/, $_)
    } @_);
}

##---------------------------------------------------------------
## $sw->pdsend(@messages)
##  + send each elt of @messages in turn to Pd, if we can.
sub pdsend {
  my $sw = shift;

  ##-- DEBUG
  #print
  #  (STDERR ref($sw), "::pdsend() called: ", @_, "\n",
  #   ">  {pd}=$sw->{pd}\n",
  #   ">  {pdsend}=$sw->{pdsend}\n",
  #   ">  {pdsendfh}=$sw->{pdsendfh}\n");
  ##--/DEBUG

  return if (!$sw->{pd});
  if (!$sw->{pdsendfh}) {
    if ($sw->{dummy}) {
      $sw->{pdsendfh} = \*STDOUT;
    }
    else {
      $sw->{pdhost} = 'localhost' if (!defined($sw->{pdhost}));
      $sw->{pdsendfh} =
	Pd::Socket::Send::TCP->new(PeerAddr=>$sw->{pdhost},
				   PeerPort=>$sw->{pdsend});
    }
  }
  if (!$sw->{pdsendfh}) {
    $sw->verr(0,"Error: could not open socket '$sw->{pdhost}:$sw->{pdsend}': $!");
    return;
  }
  $sw->{pdsendfh}->print(map { "$_;\r\n" } @_);
}


##---------------------------------------------------------------
## $rv = $sw->status($op,$rv,@args)
sub status {
 my ($sw,$op,$rv) = (shift,shift,shift);
  if ($op eq 'generate') {
    my ($sw2,$id,$str,$tree) = @_;
    &{$sw->SUPER::STATUS->{generate}}($rv,@_);
    $sw->pdsend("id $id",
		$sw->pd_sanitize_tokens2($str),
		'EOU',
	       );
  } else {
    $sw->sendosc("$op ".join(' ',@_));
  }
  $sw->pdsend('status ' . ($rv ? 1 : 0));
  return $rv;
}


###############################################################
# synth/say/save
# $rc = $sw->say_utterance($id);
sub say_utterance {
  my ($sw,$id) = @_;
  $sw->synth_wrapper('synth_utterance',$id);
  $sw->synth_wrapper('save_utterance',$id,
		     defined($sw->{fifo}) ? $sw->{fifo} : '/dev/null');
}


###############################################################
# SayWhat::TkPD
#   + combined Tk/PD mode
###############################################################
package SayWhat::TkPD;
@ISA = qw(SayWhat::PD SayWhat::Tk);

###############################################################
# Constructor: new()
sub new {
  my $that = shift;
  my $self = $that->SayWhat::PD::new(%args);
  $self->initialize(%args);
  $self->{pd} = 1;
  return $self;
}

1;
