###############################################################
#
#      File: SayWhat::Synthesizer::Rsynth.pm
#    Author: Bryan Jurish <moocow@ling.uni-potsdam.de>
# Copyright: Perl Artistic License
#
# Description: SayWhat speech synthesizer interface to Rsynth
#
###############################################################

package SayWhat::Synthesizer::Rsynth;

# system modules
use Exporter;
use Carp;
use File::Temp qw(tempfile tempdir);

# user modules
use SayWhat::Synthesizer qw(:rc :cmds);
use Speech::Rsynth;

# 3rd-party modules
use Audio::Wav;

our $VERSION = 0.01;

@ISA = qw(SayWhat::Synthesizer);

###############################################################
# Globals
###############################################################

# %DEFAULTS : default values for new() keywords
our %DEFAULTS = (
		 mode => 'wav',
		 verbose => 1,
		 dummy => 0,
		 samp_rate => 8000,
		);

our %SYNTH_DEFAULTS = (
		       # default values for Speech::Rsynth->new() keywords
		       use_audio => 0,
		       verbose => 0,
		       klatt_f0_hz => 1330,
		       mSec_per_frame => 8,
		       speed => 1,
		       dict_path => '/usr/local/lib/dict/bDict.db',
		       help_only => 0,
		      );

our $RAW2WAV_BUFSIZE = 2048;

###############################################################
# Constructor: $syn = SayWhat::Synthesizer::Rsynth->new(%args)
# + %args keys:
#   ## INHERITED
#   samp_rate => $rate,     # sampling rate for output files
#   verbose => $lvl,        # verbosity level
#   dummy => $bool,         # dummy-mode (NYI)
#   ## NEW / internal
#   synth => $rsynth,       # a Speech::Rsynth object
# + internal state:
#   utts => {
#            $uttid => {
#                       text=>$text,
#                       #mode=>$mode,
#                       #file=>$tmpfilename,
#                      },
#              ...,
#           },
#   mode => $mode,          # one of 'au'|'raw'
#   #tmpdir => $tmpdir,
###############################################################
sub new {
  my ($that,%args) = @_;
  my $class = ref($that) || $that;
  my $self = bless {}, $class;

  %$self = (utts=>{},%DEFAULTS,%args);

  ## set up guts
  unless ($self->{synth}) {
    $self->{synth} = Speech::Rsynth->new(%SYNTH_DEFAULTS,
					 samp_rate=>$self->{samp_rate});
  }
  unless ($self->{synth}) {
    carp("$class->new(): could not create Speech::Rsynth object!\n");
  }

  ## DON't always set up a tmpfile
  #unless ($self->{tmpfile}) {
  #  @$self{qw(tmpfh tmpfile)} = tempfile('sw_rsyn_XXXXXX',
  #                                       SUFFIX=>$syn->{mode},
  #				          UNLINK=>1);
  #}

  ## always set up a tmpdir
  #unless (defined($self->{tmpdir})) {
  #  $self->{tmpfir} = tempdir(
  #			      "rsXXXXXX",
  #			      #TMPDIR=>1,
  #			      #CLEANUP=>1,
  #  			     );
  #}

  return $self;
}

sub DESTROY {
  $_[0]->{synth}->Stop if ($_[0]->{synth});
}

###############################################################
# accessors
###############################################################

# %config = $syn->configure();
# $val = $syn->configure($key);
# $syn = $syn->configure(%config);
sub configure {
  my $syn = shift;
  if (scalar(@_) == 1) {
    # get value
    my $code = $syn->{synth}->can($_[0]);
    return &$code($syn->{synth}) if ($code);
    return $syn->{$_[0]};
  }


  if (scalar(@_) == 0) {
    # configure() : get whole config
    my %config = (%$syn,
		  $syn->{synth}->configure(),
		 );
    delete(@config{(qw(synth utts _uttidx),
		    qw(au_fd linear_fd dev_fd),
		    qw(au_filename linear_filename),
		    qw(jsru_name par_name),
		    qw(running use_audio flags))});
    return %config;
  }
  else {
    # configure(%args) : set config
    my ($k,$v);
    while (($k,$v) = (splice(@_,0,2))) {
      # pass through if possible
      if ($syn->{synth}->can($k) && $k ne 'verbose') {
	$syn->{synth}->$k($v);
	$syn->{$k} = $syn->{synth}->$k() if (exists($syn->{$k})); # and pass back
      } else {
	# just set locally
	$syn->{$k} = $v if (exists($syn->{$k}));
      }
    }
    return $syn;
  }
}


###############################################################
# subs: utterance maintainence
#   + $rc = add_utterance($id,$text);
#   + $rc = del_utterance($id);
#   + $rc = rename_utterance($oldid,$newid);
#   + $rc = synth_utterance($id);
#   + $rc = say_utterance($id);
#   + $rc = save_utterance($id,[$file,[$mode]]);
###############################################################

###############################################################
# add/delete/rename
sub add_utterance {
  $_[0]->{utts}{$_[1]}{text} = $_[2];
  return $RC_OK;
}
sub del_utterance {
  my ($syn,$id) = @_;
  my $utt = $syn->{utts}{$id};
  ## bad id
  if (!defined($utt)) {
    carp(ref($syn)."->del_utterance(): warning: utterance-id '$oldid' not defined!\n");
    return $RC_FAILED;
  }
  ## remove tmpfile
  #unlink($utt->{file}) if (defined($utt->{file}));
  ## delete utterance info
  delete($syn->{utts}{$id});
  return $RC_OK;
}
sub rename_utterance {
  my ($syn,$oldid,$newid) = @_;
  my $utt = $syn->{utts}{$oldid};
  if (!defined($utt)) {
    carp(ref($syn)."->rename_utterance(): warning: utterance-id '$oldid' not defined!\n");
    return $RC_FAILED;
  }
  $syn->{utts}{$newid} = $syn->{utts}{$oldid};
  delete $syn->{utts}{$oldid};
  return $RC_OK;
}

###############################################################
# synth/say
*synth_utterance = \&synth_utterance_DUMMY;
sub synth_utterance_DUMMY { return $RC_OK; }
sub synth_utterance_FILE {
  my ($syn,$id) = @_;
  my ($utt);
  ## bad id
  if (!defined($utt = $syn->{utts}{$id})) {
    carp(ref($syn)."->synth_utterance(): warning: utterance-id '$id' not defined!\n");
    return $RC_FAILED;
  }
  ## set up temp file
  $utt->{mode} = $syn->{mode} unless (defined($utt->{mode}));
  $utt->{file} = "tmp$id.$utt->{mode}" unless (defined($utt->{file}));
  ## synthesize
  return $syn->save_utterance($id,$utt->{file},$utt->{mode});
}

*say_utterance = \&say_utterance_DUMMY;
sub say_utterance_DUMMY {
  my ($syn,$id) = @_;
  my $utt = $syn->{utts}{$id};
  ## bad id
  if (!defined($utt = $syn->{utts}{$id})) {
    carp(ref($syn)."->say_utterance(): warning: utterance-id '$id' not defined!\n");
    return $RC_FAILED;
  }
  ## ... just say it
  $syn->{synth}->use_audio(1);
  return $RC_OK
    if ($syn->start
	&& ($syn->{synth}->say_string($utt->{text}) || 1)
	&& $syn->stop);
  return $RC_FAILED;
}
sub say_utterance_FILE {
  my ($syn,$id) = @_;
  my $utt = $syn->{utts}{$id};
  ## bad id
  if (!defined($utt = $syn->{utts}{$id})) {
    carp(ref($syn)."->say_utterance(): warning: utterance-id '$id' not defined!\n");
    return $RC_FAILED;
  }
  ## do we have a file?
  if (defined($utt->{file}) && -e $utt->{file}) {
    ## ... then play it
    local ($cmd);
    eval "\$cmd = \"$PLAY_CMDS{$syn->{mode}}\";";
    system("$cmd $utt->{file}");
    return $RC_OK;
  } else {
    ## ... just say it
    $syn->{synth}->use_audio(1);
    return $RC_OK
      if ($syn->start
	  && ($syn->{synth}->say_string($utt->{text}) || 1)
	  && $syn->{synth}->stop);
  }
  return $RC_FAILED;
}

sub save_utterance {
  my ($syn,$id,$file,$mode) = @_;
  ## set up filename
  $mode = $syn->{mode} unless (defined($mode));
  $file = "$id.$utt->{mode}" unless (defined($file));

  ## synthesize straight to the file
  &{$syn->filename_accessor($syn->{mode})}($syn->{synth}, $file);

  # fire 'er up
  return $RC_OK
    if ($syn->start
	&& ($syn->{synth}->say_string($syn->{utts}{$id}{text}) || 1)
	&& $syn->stop);
  return $RC_FAILED;
}


###############################################################
# subs: start/stop
###############################################################

# $syn->start();
#  + starts synth
sub start {
  return $_[0]->{synth}->start() ? $RC_OK : $RC_FAILED;
}

# $syn->stop();
#  + stops synth
sub stop {
  my $syn = shift;
  ## close down all files
  my $rc = $syn->{synth}->stop() ? $RC_OK : $RC_FAILED;
  ## post-process wav-files
  if (defined($syn->{synth}->get_linear_filename) && $syn->{mode} eq 'wav') {
    $rc = $rc && $syn->raw2wav($syn->{synth}->get_linear_filename);
  }
  ## cleanup synth flags
  $syn->{synth}->use_audio(0);
  $syn->{synth}->set_au_filename(undef);
  $syn->{synth}->set_linear_filename(undef);
  return $rc;
}

###############################################################
# subs: handy
###############################################################

# $rc = $syn->raw2wav($filename)
#  + converts filename from raw linear to riff wav (quasi-in-place)
# raw2wav($infile,$outfile,$rate,$bps)
sub raw2wav {
  my ($syn,$file) = @_;

  ## open input file
  return $RC_FAILED unless (open(RAWFILE,"<$file"));

  my $wav = Audio::Wav->new();
  return $RC_FAILED unless (defined($wav));

  my $ww = $wav->write($file.".wav", { bits_sample => 16,
				       sample_rate => $syn->{synth}->samp_rate,
				       channels => 1,
				     });
  return $RC_FAILED unless (defined($ww));

  ## read samples from input file & write them to the wav file
  my ($buf,$len);
  while ($len = read(RAWFILE,$buf,$RAW2WAV_BUFSIZE)) {
    $ww->write_raw($buf,$len);
  }

  ## cleanup
  $ww->finish();
  close(RAWFILE);

  ## shuffle
  unlink($file);
  rename($file.'.wav',$file);

  return $RC_OK;
}

# $sub = $syn->filename_accessor($mode)
sub filename_accessor {
  if ($_[0]->{mode} eq 'au') {
    return $_[0]->{synth}->can('au_filename');
  } elsif ($_[0]->{mode} eq 'raw') {
    return $_[0]->{synth}->can('linear_filename');
  } elsif ($_[0]->{mode} eq 'wav') {
    return $_[0]->{synth}->can('linear_filename');  # HACK: wav files get post-processed
  } else {
    return sub {};
  }
}


1;

__END__
